#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <regex.h>
#include "conf.h"

/* Declarations */
struct odpath_t *root;

int substitute(const char *str, char **dest, const char *var, const char *subst);
int get_next(const char *ptr, const char **dest, const char *var);
int end_of_word(char c);
int check_regex(const char *regex, const char *line);
struct odpath_t *find_path(const char *path, pathtype_t type);
struct odpath_t *find_enter(const char *path);
struct odpath_t *find_leave(const char *path);
void usage(const char *msg);

#ifdef USE_ONENTERLEAVE
void display_file(const char *path);
int exec_ok(const char *path);
#endif

/* Definitions */
int main(int argc, const char **argv) {
#ifdef USE_ONENTERLEAVE
char onenter[PATH_MAX + 1];
#endif
char working[PATH_MAX + 1], cwd[PATH_MAX + 1];
int len, i;
const char *src = NULL, *dst = NULL, *home = NULL;

	if (argc < 2)
		usage("Not enough arguments");

	for (i = 1; i < argc; ++i) {
		if (!strcmp(argv[i], "-V")) {
			printf("This is OnDir version %s\n", VERSION);
			return 0;
		}

		if (argv[i][0] == '-')
			usage("Unknown argument");

		if (!src) src = argv[i];
		else
		if (!dst) dst = argv[i];
		else
			usage("You have already specified src and dst directories");
	}

	if (!dst) {
		getcwd(cwd, PATH_MAX);
		dst = cwd;
	}

	if (src[0] != '/' || dst[0] != '/')
		fatal("either the source or destination directory is not absolute");

	root = load_conf(GLOBAL_CONF, root);

	if ((home = getenv("HOME"))) {
		snprintf(working, PATH_MAX, "%s/.ondirrc", home);
		root = load_conf(working, root);
	}

	/*
		The basic algorithm is thus:

		1. Traverse up source path until it matches the lead of dest path 
		2. Traverse from the join to the end of dest path
	*/

	strcpy(working, src);
	len = strlen(working);

	/* Traverse up source path */
	while (strncmp(working, dst, len)) {
	struct odpath_t *p = find_leave(working);

		if (p) {
		char *sub = NULL;

			if (substitute(p->content, &sub, "ONDIRWD", working) && sub != NULL)
				printf("%s", sub);
			else
				printf("%s", p->content);
			free (sub);
			sub = NULL;
		}
#ifdef USE_ONENTERLEAVE
		snprintf(onenter, PATH_MAX, "%s/.onleave", working);
		if (exec_ok(onenter))
			display_file(onenter);
#endif
		while (len && working[len] != '/')
			--len;
		working[len] = 0;
	}

	if (dst[len]) ++len;

	/* Move up to destination */
	while (dst[len] != 0) {
	struct odpath_t *p;

		while (dst[len] != 0 && dst[len] != '/')
			++len;
		strncpy(working, dst, len);
		working[len] = 0;
		if ((p = find_enter(working))) {
		char *sub = NULL;

			if (substitute(p->content, &sub, "ONDIRWD", working) && sub != NULL)
				printf("%s", sub);
			else
				printf("%s", p->content);
			free(sub);
			sub = NULL;
		}
#ifdef USE_ONENTERLEAVE
		snprintf(onenter, PATH_MAX, "%s/.onenter", working);
		if (exec_ok(onenter))
			display_file(onenter);
#endif
		if (dst[len] == '/') ++len;
	}

	return 0;
}

/* Substitue string "subst" for all occurrences of variable "var" in
 * string "str" and fill a newly allocated buffer, "dest". *
 *
 * IMPORTANT: Remember to free dest later.
 */
int substitute(const char *str, char **dest, const char *var, const char *subst) {
int len = 0;					   /* Size to be malloced. */
int retval = 0;
const char *ptmp = NULL;
const char *tmp = NULL;
char *tstr = NULL;

	if (!dest || !str || !var || !subst) {
		return (0);
	}

	/* Find length of required string. */
	ptmp = str;
	while ((retval = get_next(ptmp, &tmp, var)) != 0) {
		if (retval == -1) {
			dest = NULL;
			return (0);
		}

		len += tmp - ptmp;
		/* Leave enough space for the substitution. 1 is for $. */
		len += strlen(subst);
		ptmp = tmp + strlen(var) + 1;
		if (retval == 2) {
			ptmp += 2;					  /* Go past braces. */
		}
	}
	/* This adds the space for the part after the last variable. */
	if (ptmp) {
		/* +1 is for '\0' */
		len += strlen(ptmp) + 1;
	}

	*dest = malloc(len);
	if (*dest == NULL) {
		dest = NULL;
		return (0);
	}
	memset(*dest, 0, len);

	ptmp = str;
	tstr = *dest;
	while ((retval = get_next(ptmp, &tmp, var)) != -1) {
		/* No more variables found. tmp should be NULL here. */
		if (retval == 0) {
			/* Copy the part after the last variable. */
			strncpy (tstr, ptmp, strlen(ptmp));
			tstr += strlen(ptmp);
			*tstr = '\0';
			return 1;
			break;
		}

		/* Retval = 1 or 2. Normal variable. */
		/* Copy the string between the end of the last var and
		 * the start of this var. */
		strncpy(tstr, ptmp, tmp - ptmp);
		tstr += tmp - ptmp;

		/* Substitute for variable. */
		strcpy(tstr, subst);
		tstr += strlen(subst);

		/* Adding 1 is for '$' */
		ptmp = tmp + strlen(var) + 1;

		/* Variable of the form: ${VARIABLE}. */
		if (retval == 2) {
			/* Adding 2 is to compensate for  '{' and '}'. */
			ptmp += 2;
		}

		/* ptmp points to the end of the variable now. */
	}

	/* get_next returned -1. Error. */
	free(*dest);
	*dest = NULL;
	return 0;
}

/* Searches for the first occurrence of a variable in a given string, ptr.
 *
 * var contains something like VARNAME.
 *
 * Returns -1 on error.
 *
 * Returns 0 if there are no occurrences of the variable in ptr.
 *
 * Returns 1 if the first occurrence of the variable is in "$VARNAME"
 * (explicit end of word)
 *
 * Returns 2 if the first occurrence of the variable is in the form
 * "${VARNAME}"
 *
 */
int get_next(const char *ptr, const char **dest, const char *var) {
const char *tmp = ptr;
const char *ptmp = ptr;
static int open_brace = '{';					/* Search for this. */
static int close_brace = '}';

	if (!ptr || !var || !dest) {
		return -1;
	}
	*dest = NULL;

	while ((tmp = strchr(ptmp, '$')) != NULL) {
	int have_brace = 0;			   /* Not a braced variable */

		if (*(tmp + 1) == open_brace) {
			have_brace = 1;
			tmp++;					/* Go beyond braces. */
		}

		tmp++;						   /* Set to start of variable. */
		if (strncmp(tmp, var, strlen (var)) == 0) {
			/* Catch: Var with starting brace, but no ending one. */
			if (have_brace == 1 && *(tmp + strlen(var)) != close_brace) {
				ptmp = tmp;
				continue;
			}

			/* Catch: Var not ending at word (like $ONDIRWDDD) */
			if (have_brace == 0 && !end_of_word(*(tmp + strlen(var)))) {
				ptmp = tmp;
				continue;
			}

			*dest = tmp - 1 - have_brace;
			return (have_brace + 1);
		}

		ptmp = tmp;
	}

	*dest = NULL;
	return 0;
}

/* Checks if c is an end-of-word character (like ' ', '\0', ...)
 *
 * Returns:
 * 1 if it is.
 * 0 if it's not.
 * */
int end_of_word (char c)
{
/* Characters which are considered to end a word. */
char *end_chars = " \n\t{}[]()\"";

	if (c == '\0') {
		return 1;
	}

	while (*end_chars) {
		if (c == *end_chars) {
			return 1;
		}
		end_chars++;
	}

	return 0;
}

/* Check if a given regex matches a given line.
 *
 * Returns:
 * 0 if they don't match.
 * 1 if they match.
 */
int check_regex(const char *regex, const char *line) {
regex_t preg;
regmatch_t match = { 0, 0 };
int retval = 0;

	if (!regex || !line) {
		/* FAILURE */
		return (0);
	}

	if ((retval = regcomp(&preg, regex, REG_EXTENDED))) {
	char err_buf[512];

		regerror(retval, &preg, err_buf, 512);
		fprintf(stderr, "ondir: regcomp failed, %s\n", err_buf);
		return 0;
	}

	retval = regexec(&preg, line, 1, &match, 0);
	regfree(&preg);

	return retval == 0 && match.rm_so == 0 && match.rm_eo == strlen(line);
}

struct odpath_t *find_path(const char *path, pathtype_t type) {
struct odpath_t *i;
int j;

	for (i = root; i != NULL; i = i->next)
		if (i->type == type)
			for (j = 0; j < i->npaths; ++j)
				if (!strcmp(i->paths[j], path) || check_regex(i->paths[j], path))
					return i;
	return NULL;
}

#ifdef USE_ONENTERLEAVE
void display_file(const char *path) {
FILE *fp;

	if ((fp = fopen(path, "r"))) {
	char buffer[512];
	int len;

		while ((len = fread(buffer, 1, 512, fp)) > 0)
			fwrite(buffer, len, 1, stdout);
		fclose(fp);
	}
}

int exec_ok(const char *path) {
struct stat s;

	return
		stat(path, &s) == 0 && /* exists.. */
		s.st_mode & (S_IRUSR | S_IRGRP | S_IROTH) && /* ...readable... */
		s.st_uid == getuid() && /* ...owned by me... */
		!(s.st_mode & (S_IWGRP | S_IWOTH)); /* ...not group/world writeable */
}
#endif

struct odpath_t *find_enter(const char *path) {
	return find_path(path, PT_ENTER);
}

struct odpath_t *find_leave(const char *path) {
	return find_path(path, PT_LEAVE);
}

void usage(const char *msg) {
	printf(
		"%s\n"
		"usage: ondir [-V] <old-directory> [<new-directory>]\n"
		"\n"
		"  -V    Displays version information.\n"
		"\n"
		"  <old-directory> is the last working directory.\n"
		"\n"
		"  <new-directory> is the current working directory. If <new-directory> is\n"
		"  omitted, the current working directory is obtained via a call to getcwd().\n"
		"\n", msg
	);
	exit(1);
}

